/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <iostream>
#include <vector>
#include <list>

#include "./AsyncQuery.h"

namespace hw4 {

// The maximum number of open (FILE *)s we will tolerate, and hence
// the maximum number of index files we can support.
static const int kMaxFILE = 64;

// This is the thread initialization function that our worker threads call
// when they are born.  Threads use it to pick up their own QueryProcessor.
void *ThreadPickupQP(void *v) {
  AsyncQuery *aq = static_cast<AsyncQuery *>(v);
  return static_cast<void *>(aq->GetQP());
}

// This is the thread dispatch function that we use; when a worker
// thread picks up a Task, this is the function that the Task directs
// the worker thread to invoke.
void ThreadDispatchQP(ThreadPool::Task *arg) {
  // "arg" is the AsyncQuery::QueryTask created for this query.  It
  // contains a number of fields the worker thread needs in order to
  // get its work done.
  AsyncQuery::QueryTask *qt = (AsyncQuery::QueryTask *) arg;

  // Pull the fields we need to use out of the AsyncQuery::QueryTask.
  AsyncQuery *aq = qt->aq;
  int writepipe_fd = qt->writepipe_fd;
  hw3::QueryProcessor *qp = (hw3::QueryProcessor *) qt->thread_init_return;

  // Invoke the query on this thread's QueryProcessor and store the
  // results in the "res" field of the QueryTask.
  qt->res = qp->ProcessQuery(qt->query);

  // Enqueue the query result so that the customer can pick it up.
  aq->EnqueueResults(qt);

  // Signal the writepipe to indicate to the customer that a query has
  // completed.
  char c = 0xFF;
  while (1) {
    int res = write(writepipe_fd, &c, 1);
    if (res == 1) {
      break;
    }
    if (res == -1) {
      if (errno == EINTR)
        continue;
    }
    assert(0);  // A pipe error happened; we won't try to deal with it.
  }
}

AsyncQuery::AsyncQuery(list<string> indexfile_names, int *retfd) {
  // Create the query processors, one per worker thread, and stash
  // them away in the qp_ member variable.  We calculate the number
  // of worker threads (and query processors) in a way that limits
  // the number of open (FILE *) streams to kMaxFILE.  Note that
  // each query processor opens one (FILE *) per indexfile.
  assert(indexfile_names.size() > 0);
  numqp_ = kMaxFILE / indexfile_names.size();
  assert(numqp_ > 0);
  qp_ = new hw3::QueryProcessor*[numqp_];
  std::cout << "  validating index files' checksums..." << std::endl;
  for (int i = 0; i < numqp_; i++) {
    // Only validate the index files during the first loop iteration,
    // since checksum calculation is relatively slow.
    qp_[i] = new hw3::QueryProcessor(indexfile_names,
                                     i == 0 ? true : false);
  }

  // Initialize the task completion queue mutex.
  assert(pthread_mutex_init(&qlock_, NULL) == 0);

  // Initialize the selfpipe.
  assert(pipe(&(selfpipe_[0])) == 0);
  int flags = 0;
  flags = fcntl(selfpipe_[0], F_GETFL, NULL);
  assert(flags != -1);
  assert(fcntl(selfpipe_[0], F_SETFL, flags | O_NONBLOCK) == 0);

  // Return the read end of the selfpipe to the caller.
  *retfd = selfpipe_[0];

  // Initialize the thread pool.
  threadnum_ = 0;
  tp_ = new ThreadPool(numqp_,
                       ThreadPickupQP,
                       static_cast<void *>(this));
}

AsyncQuery::~AsyncQuery() {
  // Destroy the threadpool, if it hasn't already happened.
  ShutdownThreadPool();

  // Destroy the QueryProcessors and then the qp_ array itself.
  for (int i = 0; i < numqp_; i++)
    delete qp_[i];
  delete[] qp_;

  // Shut down the selfpipe.
  close(selfpipe_[0]);
  close(selfpipe_[1]);
}

void AsyncQuery::ProcessQuery(vector<string>query, void *v) {
  // Create a QueryTask object, with the thread dispatch function set
  // to be ThreadDispatchQP.
  AsyncQuery::QueryTask *qt =
    new AsyncQuery::QueryTask(ThreadDispatchQP);
  // Populate the rest of the fields of the QueryTask object.
  qt->v = v;
  qt->aq = this;
  qt->writepipe_fd = selfpipe_[1];
  qt->query = query;

  // Dispatch the AsyncQuery::QueryTask to the thread pool.
  tp_->Dispatch(qt);
}

// Pick up query results and return them, or return NULL if there
// aren't any.
AsyncQuery::QueryTask *AsyncQuery::DequeueResults() {
  // Grab the lock, dequeue the next result.
  AsyncQuery::QueryTask *retTask = NULL;
  pthread_mutex_lock(&qlock_);
  if (result_queue_.size() > 0) {
    retTask = result_queue_.front();
    result_queue_.pop_front();
  }
  pthread_mutex_unlock(&qlock_);
  return retTask;
}

// A worker thread calls this to queue a completed task.
void AsyncQuery::EnqueueResults(QueryTask *t) {
  // Grab the lock, queue the result.
  pthread_mutex_lock(&qlock_);
  result_queue_.push_back(t);
  pthread_mutex_unlock(&qlock_);
}

// A newly-born thread pool worker calls this during initialization to
// pick up its QueryProcessor.
hw3::QueryProcessor *AsyncQuery::GetQP() {
  hw3::QueryProcessor *retqp = qp_[threadnum_];
  threadnum_++;
  return retqp;
}

void AsyncQuery::ShutdownThreadPool() {
  // Delete the thread pool, if it is not already NULL. The threadpool
  // destructor tells the worker threads to kill themselves and also
  // cleans up any pending tasks.
  if (tp_ != NULL) {
    delete tp_;
    tp_ = NULL;
  }
}

}  // namespace hw4
