/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _HW4_ASYNCQUERY_H_
#define _HW4_ASYNCQUERY_H_

extern "C" {
#include <pthread.h>                  // for pthread mutexes
}

#include <list>                       // for std::list
#include <string>                     // for std::string
#include <vector>                     // for std::vector

#include "./ThreadPool.h"             // for hw4::ThreadPool
#include "./libhw3/QueryProcessor.h"  // for hw3::QueryProcessor

using std::list;

namespace hw4 {

// The AsyncQuery class abstracts away the asynchronous processing
// of queries against a set of index files.  Internally, the
// AsyncQuery creates a set of hw3::QueryProcessor objects and a
// hw4::ThreadPool.  There is a one-to-one mapping from
// QueryProcessors to threads in the pool; since many UNIX'en have
// a (surprisingly small) limit to the number of simultaneously open
// (FILE*) streams, we cap the ThreadPool to a small number of
// threads.
//
// As well, the AsyncQuery class opens a self-pipe.  When a query
// is complete, thread that processed the query writes a byte on the
// write-end of the self-pipe so that a customer doing a select() on
// the read-end will be woken up.  (The read-end file descriptor is
// set to O_NONBLOCK.)
//
// The AsyncQuery maintains a FIFO completion queue from which
// customers can pick up completed queries.
class AsyncQuery {
 public:
  // Construct an AsyncQuery object.  "indexfile_names" is a list of
  // indexfile names and "retfd" is an output parameter through which
  // the file descriptor of the read-end of the self-pipe is returned.
  // Customers should select() or poll() on *retfd to notice when new
  // results are ready to be dequeued.
  AsyncQuery(list<string> indexfile_names, int *retfd);
  virtual ~AsyncQuery();

  // This inner class defines what a QueryTask is.  (Customers don't
  // populate this structure; it is populated as a side-effect of a
  // customer invoking ProcessQuery().)  The "query" element of the
  // task is a list of strings, and customers provide the "v"
  // argument to stash away a pointer to anything they  see fit.
  //
  // When the query completes, the "res" field is filled in by the
  // worker thread and the AsyncQuery::QueryTask object is enqueued
  // on the output queue.  Customers can dequeue the task using the
  // "DequeueResults()" method.
  class QueryTask : public ThreadPool::Task {
   public:
    explicit QueryTask(ThreadPool::thread_task_fn f) : Task(f) { }

    // Customers use "v" as they see fit; its contents will be
    // unperturbed and available to the customer when they pick up
    // their result on the output queue.
    void *v;

    // This field stores a pointer to the AsyncQuery to which a worker
    // thread should give the query results (via the Enqueue() method).
    AsyncQuery *aq;

    // This field stores the write-end of the self-pipe used to signal
    // the availability of a new query result to a customer.
    int writepipe_fd;

    // Customers pass in a query to ProcessQuery(), which stores the
    // query in this field. A query contains a vector of query words.
    vector<string> query;

    // When a query has been processed, the worker thread will store
    // the query results in this field.  When a customer picks up a
    // QueryTask, they should read this field to get their results.
    vector<hw3::QueryProcessor::QueryResult> res;
  };

  // Customers use this to enqueue an asynchronous query.  AsyncQuery
  // allocates a QueryTask to store the query; the customer can pass
  // in "v",  which is stashed in the allocated QueryTask and
  // available on completion.  A customer can pick up a completed
  // query by calling DequeueResults().  When the query completes, the
  // query processor thread will write one byte on selfpipe_[1] to
  // notify the caller.
  void ProcessQuery(vector<string> query, void *v);

  // Customers call this to dequeue their query results, if they are
  // available.  Returns NULL if there are no results, or a filled in
  // (AsyncQuery::QueryTask *) if there are some. The caller takes
  // ownership of the returned (AsyncQuery::QueryTask *) and must
  // delete it.
  QueryTask *DequeueResults();

  // The thread pool worker threads that are processing a query call
  // this function to enqueue query results.  Customers should not use
  // this function.
  void EnqueueResults(QueryTask *t);

  // Worker threads call this function during initialization to pick
  // up their QueryProcessor.  Customers should not use this function.
  hw3::QueryProcessor *GetQP();

  // A customer can call this during the shutdown procedure to tell
  // the AsyncQuery to clear out its threadpool.
  void ShutdownThreadPool();

 private:
  // The queue of completed queries and a mutex to protect it.
  list<QueryTask *>    result_queue_;
  pthread_mutex_t      qlock_;

  // The ThreadPool of worker threads;
  ThreadPool          *tp_;

  // The set of QueryProcessors, one per thread in the pool.
  hw3::QueryProcessor **qp_;

  // The ends of the notification self-pipe.  0 is the read end, 1 is
  // the write end.
  int selfpipe_[2];

  // This field is used as an index into qp_ by worker threads in
  // their thread initialization function; they use this to pick up
  // the appropriate (QueryProcessor *) from qp_[threadnum_].
  int threadnum_;

  // The number of QueryProcessors in qp_; this is also the number of
  // worker threads in the ThreadPool tp_.
  int numqp_;
};

}  // namespace hw4

#endif  // _HW4_ASYNCQUERY_H_
