/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <boost/algorithm/string.hpp>
#include <string.h>
#include <algorithm>
#include <iostream>
#include <string>

#include "./SocketUtils.h"
#include "./HttpConnection.h"

namespace hw4 {

HttpConnection::HttpConnection(int client_fd, EventLoop *looper,
                               RequestProcessor *rp)
  : client_fd_(client_fd), loop_(looper), rp_(rp), slr_(client_fd) {
  // An HttpConnection begins life in the "reading a request" state.
  state_ = READING_REQUEST;

  // Register ourselves with the event loop.  Since we're starting
  // in the reading request state, we'll only request "socket is
  // readable" or "socket has failed" events.
  assert(loop_->RegisterFD(client_fd_,
                           EventLoop::FD_IS_READABLE |
                           EventLoop::FD_ERR |
                           EventLoop::FD_HUP,
                           this));
}

HttpConnection::~HttpConnection() {
  // We're being destroyed!  Close the listening socket, if it hasn't
  // already been closed yet.
  if (client_fd_ != 0) {
    close(client_fd_);
    client_fd_ = 0;
  }
}

void HttpConnection::Handle(const int fd,
                        const unsigned int eventfield,
                        EventLoop *looper) {
  assert(fd == client_fd_);

  if (state_ == READING_REQUEST) {
    // We're in the "reading from a request" state, so invoke that
    // event handler.
    ProcessRead(eventfield);
    return;
  } else if (state_ == GENERATING_REPLY) {
    std::cerr << "Should never get Connection::Handle() event ";
    std::cerr << "in the GENERATING_REPLY state.  Aborting.";
    std::cerr << std::endl;
    assert(0);
  } else if (state_ == WRITING_REPLY) {
    // We're in the "writing a reply" state, so invoke that
    // event handler.
    ProcessWrite(eventfield);
    return;
  }

  std::cerr << "A client connection is in a bogus state?!." << std::endl;
  assert(0);
}

void HttpConnection::ProcessRead(const unsigned int eventfield) {
  // Handle the "socket is readable" or "socket is in an error state"
  // event. The steps, roughly, are:
  //
  //   (1) make sure we received the right kind of event
  //   (2) use the SocketLineReader slr_ to process more data
  //       from our socket
  //   (3) test to see if we spotted the end of the HTTP request.
  //       we'll know because we see an empty line (i.e, a
  //       line containing nothing but "\r\n") at some point in
  //       the lines_ vector<string>.
  //   (4) use "GenerateHeaders()" to process the request.
  //   (5) transition into the GENERATING_REPLY state and invoke
  //       the RequestProcessor rp_'s "Process()" function to
  //       process the request.

  // MISSING
}

bool HttpConnection::GenerateHeaders() {
  // Erase the current set of headers, if stored.
  headers_.clear();

  // Try to extract the HTTP request method and url from the first
  // line.  We only handle "GET" requests in this simple server,
  // so return false if it doesn't look like one.
  vector<string> strs;
  boost::split(strs, lines_[0], boost::is_any_of(" \t\r\n"));
  if ((strs.size() < 2) || (!(strs[0] == "GET")))
    return false;

  // Pull out the URL.
  headers_["get"] = strs[1];

  // Try to extract the rest of the headers.
  for (unsigned int i = 1; i < lines_.size(); i++) {
    if (lines_[i] == "\r\n")
      break;
    boost::split(strs, lines_[i], boost::is_any_of(" \t\r\n"));
    if (strs.size() >= 2) {
      // Extract out the header name, lowercase it, strip the ":".
      std::transform(strs[0].begin(), strs[0].end(), strs[0].begin(),
                     ::tolower);
      if ((strs[0].at(strs[0].size()-1)) == ':') {
        strs[0].erase(strs[0].size()-1);
      }

      // Extract out the header value, lowercase it.
      std::transform(strs[1].begin(), strs[1].end(), strs[1].begin(),
                     ::tolower);
      headers_[strs[0]] = strs[1];
    }
  }

  // Return true to signal we found a GET request, at least.
  return true;
}

void HttpConnection::Quitting(int fd) {
  // Don't need to do anything to handle this event; our destructor
  // is enough.
}

void HttpConnection::Response(string response) {
  // The RequestProcessor is handing us a response.  Stash it away.
  response_ = response;
  response_written_ = 0;

  // Transition to the WRITING_REPLY state and register interest in
  // writeable events.
  state_ = WRITING_REPLY;
  loop_->RegisterFD(client_fd_,
                    EventLoop::FD_IS_WRITEABLE |
                    EventLoop::FD_ERR |
                    EventLoop::FD_HUP,
                    this);
}

void HttpConnection::ProcessWrite(const unsigned int eventfield) {
  // Make sure we got a "writeable" or "error" event.
  assert(!(eventfield & EventLoop::FD_IS_READABLE));
  if ((eventfield & EventLoop::FD_ERR) ||
      (eventfield & EventLoop::FD_HUP)) {
    // Drop the client, since we just got a socket error.
    loop_->UnregisterFD(client_fd_);
    return;
  }

  // Write back as much data as we can.
  uint32_t amount_written = 0;
  uint32_t len = response_.length();
  const char *data = response_.data();
  if (!NBWrite(client_fd_,
               data+response_written_,
               len - response_written_,
               &amount_written)) {
    // We hit EOF, so unregister ourselves.
    loop_->UnregisterFD(client_fd_);
    return;
  }

  // We wrote some more data.  See if we're done, and if so,
  // close the client connection.
  response_written_ += amount_written;
  if (response_written_ >= len) {
    // We finished writing!  Unregister ourselves.
    loop_->UnregisterFD(client_fd_);
  }

  // Exit the event handler.
}

}  // namespace hw4
