/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _HW4_HTTPCONNECTION_H_
#define _HW4_HTTPCONNECTION_H_

#include <stdint.h>
#include <map>
#include <string>
#include <vector>

#include "./EventLoop.h"
#include "./EventLoopHandler.h"
#include "./RequestProcessor.h"
#include "./SocketLineReader.h"

using std::map;
using std::string;
using std::vector;

namespace hw4 {

class RequestProcessor;
class EventLoop;

// An HttpConnection class manages a new http client connection.
// When a connection becomes available, the Listener class
// manufactures a new HttpConnection object and passes it the client
// socket descriptor.  The HttpConnection registers itself as an
// EventLoopHandler with the event loop, reading and writing data to
// process requests.
//
// When the connection is finished, the HttpConnection class
// unregisters itself from the EventLoop and closes the client socket.
class HttpConnection : public EventLoopHandler {
 public:
  HttpConnection(int client_fd, EventLoop *looper, RequestProcessor *rp);
  virtual ~HttpConnection();

  // This is the EventLoopHandler function that the EventLoop invokes
  // on us when a socket is ready to write or when it experiences
  // some kind of error.
  virtual void Handle(const int fd,
                      const unsigned int eventfield,
                      EventLoop *looper);

  // This is the EventLoopHandler function that the EventLoop invokes
  // on us when it is shutting down.
  virtual void Quitting(int fd);

  // A response generator invokes this function on us to hand us
  // response HTTP+HTML data to send back to the client.
  virtual void Response(string response);

 private:
  // This enum defines the states that this connection could be in.
  enum ConnectionState {
    READING_REQUEST,   // We're reading the HTTP request from the client.
    GENERATING_REPLY,  // We're generating an HTTP response.
    WRITING_REPLY      // We're writing the response to the client.
  };

  // This variable tracks the current state of this connection.
  ConnectionState state_;

  // This is the file descriptor associated with the client socket.
  int client_fd_;

  // This is the EventLoop on which we registered ourselves as a
  // handler for all of the open client sockets.
  EventLoop *loop_;

  // This is the RequestProcessor we should invoke to process
  // requests.
  RequestProcessor *rp_;

  // This is our SocketLineReader helper object.
  SocketLineReader slr_;

  // These are the request header lines we've received so far.
  vector<string> lines_;

  // Once all headers have been read, we produce a map from header
  // name to header value. This is that parsed header map.
  map<string, string> headers_;

  // This is the response to return; when a response processor has
  // produced it, it will invoke Response(), which stashes the response
  // here.
  string response_;

  // This is how much of the response we've written so far to clients.
  uint32_t response_written_;

  // A private helper function to handle "socket is readable" events.
  void ProcessRead(const unsigned int eventfield);

  // A private helper function to extract the HTTP header map from
  // the set of header lines stored in lines_.
  bool GenerateHeaders();

  // A private helper function to handle "socket is writeable" events.
  void ProcessWrite(const unsigned int eventfield);
};  // class HttpConnection

}  // namespace hw4

#endif  // _HW4_HTTPCONNECTION_H_
