/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <locale>
#include <errno.h>
#include <algorithm>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include "./HttpUtils.h"
#include "./RequestProcessor.h"

using std::string;
using std::stringstream;
using std::vector;

namespace hw4 {

// QueryHttpHtml() generates a query submission form, and if qt is
// non-NULL, it also includes the previous query and query results in
// the generated HTML.
void QueryHttpHtml(HttpConnection *connection, AsyncQuery::QueryTask *qt);

RequestProcessor::RequestProcessor(EventLoop *looper,
                                   list<string> indexfiles)
  : aq_(indexfiles, &asyncquery_read_fd_) {
  // We created the AsyncQuery object in our initialization list.
  // Now, register the file descriptor that was handed back to us with
  // the event loop.
  loop_ = looper;
  assert(loop_->RegisterFD(asyncquery_read_fd_,
                           EventLoop::FD_IS_READABLE |
                           EventLoop::FD_ERR |
                           EventLoop::FD_HUP,
                           this));
}

RequestProcessor::~RequestProcessor() {
  // Shutdown the AsyncQuery object's thread pool.  This forces all
  // remaining tasks on its thread pool queue to be processed, placing
  // the remaining of our allocated structures on the AsyncQuery
  // completion queue for us to pick up.
  aq_.ShutdownThreadPool();

  // Pick up remaining tasks on the AsyncQuery completion queue and
  // delete them.
  AsyncQuery::QueryTask *nt = aq_.DequeueResults();
  while (nt != NULL) {
    delete reinterpret_cast<QueryRequest *>(nt->v);
    delete nt;
    nt = aq_.DequeueResults();
  }
}

void RequestProcessor::Process(map<string, string> headers,
                               HttpConnection *connection) {
  // Did the client ask us to exit the web server?
  // (Please leave this exit URL in place; we'll need it to
  // test that your web server has no memory leaks.)
  if (headers["get"] == string("/secret/quit/url")) {
    loop_->Quit();
    return;
  }

  // Extract out the arguments from the URL.
  URLParser p(headers["get"]);

  // Is this a HW3 query?
  if (p.get_path() == "/query") {
    if ((p.get_args().size() > 0) &&
        (p.get_args()[0].first == "terms") &&
        (p.get_args()[0].second.size() > 0)) {
      // We have a query to process.  Convert the query terms
      // into a vector of strings.
      string qstring = p.get_args()[0].second;
      std::transform(qstring.begin(), qstring.end(), qstring.begin(),
                     ::tolower);
      vector<string> terms;
      boost::split(terms, qstring, boost::is_any_of(" "));

      // Launch the query.
      QueryRequest *q = new QueryRequest();
      q->connection = connection;
      q->query = terms;
      aq_.ProcessQuery(terms, reinterpret_cast<void *>(q));
      return;
    }
  }

  // Is this a file request?
  if (boost::starts_with(p.get_path(), "/file/")) {
    connection->Response("FILE PROCESSING NOT YET IMPLEMENTED.");
    return;
  }

  // This is not a URL that we recognize; return the main query splash
  // page.
  QueryHttpHtml(connection, NULL);
}

void RequestProcessor::Quitting(int fd) { }

void RequestProcessor::Handle(const int fd,
                              const unsigned int eventfield,
                              EventLoop *looper) {
  // Is this a HW3 query completion?
  if (fd == asyncquery_read_fd_) {
    // Yes.
    HandleAsyncQuery(fd, eventfield, looper);
  } else {
    assert(0);
  }
}

void RequestProcessor::HandleAsyncQuery(const int async_fd,
                                        const unsigned int eventfield,
                                        EventLoop *looper) {
  // Read from the asyncquery fd, since some responses are supposedly
  // finished.  We'll drain the self-pipe, processing all responses.
  while (1) {
    int c;
    ssize_t res = read(async_fd, &c, 1);
    if (res == 1) {
      // Got a result; pick it up from the AsyncQuery queue.
      AsyncQuery::QueryTask *t = aq_.DequeueResults();
      assert(t != NULL);

      // Extract and return the results.
      QueryRequest *res = reinterpret_cast<QueryRequest *>(t->v);
      assert(res != NULL);

      // Call QueryHttpHtml() to generate an HTML response.
      QueryHttpHtml(res->connection, t);
      delete t;
      delete res;
    } else if (res == 0) {
      // This should never happen.
      std::cerr << "Hit EOF on self-pipe, but that shouldn't happen.";
      std::cerr << std::endl;
      assert(0);
    } else if (res == -1) {
      if ((errno == EAGAIN) || (errno == EWOULDBLOCK)) {
        // all done; exit out.
        break;
      }
      // This should never happen.
      std::cerr << "Error on self-pipe, but that shouldn't happen.";
      std::cerr << std::endl;
      assert(0);
    }
  }
}

void QueryHttpHtml(HttpConnection *connection, AsyncQuery::QueryTask *qt) {
  // If "qt" is NULL, there was no query.  If it is non-NULL,
  // it contains a query and the query results.
  //
  // Finish generating the HTTP response and also some HTML.
  // The HTML should include a FORM that submits a query using
  // the HTTP "GET" method.  If "qt" is non-NULL, you should display
  // query results after your form.
  //
  // HINT: try running our http333d server and looking at the HTML
  // that gets generated when a user visits it, and after a user
  // submits a query.  Mimic that.
  stringstream ss;
  ss << "HTTP/1.1 200 OK\r\n";

  // MISSING


  ss << "</body></html>\n";
  connection->Response(ss.str());
}

}  // namespace hw4
