/*
 * Copyright 2011 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _HW4_REQUESTPROCESSOR_H_
#define _HW4_REQUESTPROCESSOR_H_

#include <list>
#include <map>
#include <string>
#include <utility>
#include <vector>

#include "./AsyncQuery.h"
#include "./EventLoop.h"
#include "./EventLoopHandler.h"
#include "./HttpConnection.h"

using std::list;
using std::map;
using std::pair;
using std::string;

namespace hw4 {

// A RequestProcessor manages the asynchronous processing of HTTP
// requests.  First, the RequestProcessor parses the request's URL.
// If is a query submission, the RequestProcessor uses its AsyncQuery
// object to process the query.  If it is a file request, the
// RequestProcessor uses its AsyncFile object to process the request.
//
// When the RequestProcessor is constructed, it creates an AsyncQuery
// objec.  From this, it receives the read-end of a self-pipe.
// RequestProcessor registers itself with the EventLoop as a handler
// for read events on this self-pipe.
//
// HttpConnection objects invoke "Process()" to pass in a new
// request to process.

// Forward declarations.
class EventLoop;
class HttpConnection;

class RequestProcessor : public EventLoopHandler {
 public:
  // "looper" is the eventloop we should register ourselves with,
  // and "indexfiles" is a list of index files we process against.
  RequestProcessor(EventLoop *looper, list<string> indexfiles);
  virtual ~RequestProcessor();

  // An HttpConnection object will invoke this method to tell us
  // to begin processing a new URL.  (The URL is the the value stored
  // in headers["get"].)  When we are done, we invoke the
  // HttpConnection object's Results() method with our response.
  virtual void Process(map<string, string> headers,
                       HttpConnection *connection);

  // This is the EventLoopHandler function that the EventLoop invokes
  // on us when it is shutting down.
  virtual void Quitting(int fd);

  // This is the handler that the event loop invokes on us to
  // tell us that our read-end of the self-pipe has some kind
  // of readable or error event to report.
  virtual void Handle(const int fd,
                      const unsigned int eventfield,
                      EventLoop *looper);

 private:
  // This private function is called by Handle() in order to process
  // an AsyncQuery completion.
  void HandleAsyncQuery(const int async_fd,
                        const unsigned int eventfield,
                        EventLoop *looper);

  // This is the fd associated with our read-end of the AsyncQuery
  // read pipe.
  int asyncquery_read_fd_;

  // This is the AsyncQuery object we've stashed away to process
  // query requests asynchronously.
  AsyncQuery aq_;

  // This is the EventLoop on which we registered ourselves as a
  // handler for the read-end of the self pipe.
  EventLoop *loop_;
};  // class RequestProcessor

// This structure is what we pass in as the "v" argument to
// AsyncQuery::ProcessQuery().
typedef struct {
  HttpConnection *connection;
  vector<string> query;
} QueryRequest;

// This structure is what we pass in as the "v" argument to
// AsyncFile::ProcessFile().
typedef struct {
  HttpConnection *connection;
  string filename;
} FileRequest;

}  // namespace hw4

#endif  // _HW4_REQUESTPROCESSOR_H_
